/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   KfField.cxx
/// \brief  Magnetic field representation in KF (source)
/// \author Sergei Zharko <s.zharko@gsi.de>
/// \since  22.07.2024

#include "KfField.h"

#include <mutex>
#include <sstream>
#include <tuple>

using cbm::algo::kf::EFieldMode;
using cbm::algo::kf::Field;
using cbm::algo::kf::FieldFactory;
using cbm::algo::kf::detail::FieldBase;

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T, EFieldMode FldMode>
std::string Field<T, FldMode>::ToString(int indentLevel, int verbose) const
{
  constexpr char IndentChar = '\t';
  std::stringstream msg;
  std::string indent(indentLevel, IndentChar);
  msg << indent << "Field near primary vertex:\n" << this->fPrimVertexField.ToString(indentLevel + 1, verbose) << '\n';
  msg << indent << "Field type: " << static_cast<int>(this->fFieldType) << '\n';
  if constexpr (FldMode == EFieldMode::Orig) {
    //msg << indent << "Original field function";
  }
  else if constexpr (FldMode == EFieldMode::Intrpl) {
    msg << indent << "Field slices:";
    for (const auto& fldSlice : this->fvFieldSlices) {
      msg << indent << "\n - " << fldSlice.ToString(indentLevel + 1, verbose);
    }
  }
  return msg.str();
}

namespace cbm::algo::kf
{
  template class Field<float, EFieldMode::Orig>;
  template class Field<float, EFieldMode::Intrpl>;
  template class Field<double, EFieldMode::Orig>;
  template class Field<double, EFieldMode::Intrpl>;
  template class Field<fvec, EFieldMode::Orig>;
  template class Field<fvec, EFieldMode::Intrpl>;
}  // namespace cbm::algo::kf


// ---------------------------------------------------------------------------------------------------------------------
//
void FieldFactory::AddSliceReference(double halfSizeX, double halfSizeY, double zRef)
{
  if (!fSliceReferences.emplace(halfSizeX, halfSizeY, zRef).second) {
    std::stringstream msg;
    msg << "FieldFactory::AddReference: attempt of adding another slice reference with zRef = " << zRef
        << "(halfSizeX = " << halfSizeX << ", halfSizeY = " << halfSizeY << ").\nThe next slice references were "
        << "added:";
    for (const auto& el : fSliceReferences) {
      msg << "\n\t- halfSizeX = " << el.fHalfSizeX << ", halfSizeY = " << el.fHalfSizeY << ", zRef = " << el.fRefZ;
    }
    throw std::logic_error(msg.str());
  }
}

