/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   KfField.h
/// \brief  Magnetic field representation in KF (header)
/// \author Sergei Zharko <s.zharko@gsi.de>
/// \since  22.07.2024

#pragma once

#include "KfDefs.h"
#include "KfFieldRegion.h"
#include "KfFieldSlice.h"

#include <array>
#include <set>

namespace cbm::algo::kf
{
  namespace detail
  {
    /// \class  FieldBase
    /// \brief  Data members of the Field class (primary template)
    /// \tparam T        Underlying floating point type (float/double/fvec)
    /// \tparam FldMode  Field type, defined by EFieldMode
    template<typename T, EFieldMode FldMode>
    class alignas(VcMemAlign) FieldBase {
    };

    /// \class  FieldBase<T, Orig>
    /// \brief  FieldBase implementation for the original field
    /// \tparam T  Underlying floating point type (float/double/fvec)
    template<typename T>
    class alignas(VcMemAlign) FieldBase<T, EFieldMode::Orig> {
      template<typename, EFieldMode>
      friend class FieldBase;

     public:
      using FieldRegion_t = FieldRegion<T, EFieldMode::Orig>;
      using FieldValue_t  = FieldValue<T>;

      /// \brief Sets magnetic field function
      /// \param fieldFn  Magnetic field function (KF-format)
      void SetFieldFunction(const FieldFn_t& fieldFn) { fFieldFn = fieldFn; }

      /// \brief Creates field region object
      FieldRegion_t MakeFieldRegion() const { return FieldRegion_t(fFieldType, fFieldFn); }

     protected:
      /// \brief Default constructor
      FieldBase() = default;

      /// \brief  Copy constructor
      /// \tparam I  Underlying floating type of the source
      template<typename I>
      FieldBase(const FieldBase<I, EFieldMode::Orig>& other) : fFieldFn(other.fFieldFn)
      {
      }

      /// \brief  Copy assignment operator
      FieldBase& operator=(const FieldBase& other) = default;

      FieldFn_t fFieldFn{defs::ZeroFieldFn};    ///< Field function: (x,y,z) [cm] -> (Bx,By,Bz) [kG]
      EFieldType fFieldType{EFieldType::Null};  ///< Field type

     private:
      /// Serialization function
      friend class boost::serialization::access;
      template<class Archive>
      void serialize(Archive& ar, const unsigned int)
      {
        ar& fFieldType;
      }
    };


    /// \class  FieldBase<T, EFieldMode::Intrpl>
    /// \brief  Data members of the Field class with the interpolation of the magnetic field
    /// \tparam T  Underlying floating point type (float/double/fvec)
    template<typename T>
    class alignas(VcMemAlign) FieldBase<T, EFieldMode::Intrpl> {
      template<typename, EFieldMode>
      friend class FieldBase;

      using SlicesContainer_t = std::vector<FieldSlice<T>>;

     public:
      using FieldRegion_t = FieldRegion<T, EFieldMode::Intrpl>;
      using FieldValue_t  = FieldValue<T>;

      /// \brief Sets a field slice
      /// \param fieldSlice  Field slice object
      void AddFieldSlice(const FieldSlice<T>& fieldSlice) { fvFieldSlices.push_back(fieldSlice); }

      /// \brief Creates field value object
      /// \param sliceID  Index of slice
      /// \param x        x-coordinate of the point [cm]
      /// \param y        y-coordinate of the point [cm]
      FieldValue_t GetFieldValue(int sliceID, const T& x, const T& y) const
      {
        return fvFieldSlices[sliceID].GetFieldValue(x, y);
      }

      /// \brief Gets number of field slices in the instance
      int GetNofFieldSlices() const { return fvFieldSlices.size(); }

      /// \brief Creates field region object
      /// \param b0  Field value in the first node [kG]
      /// \param z0  First node z-coordinate [cm]
      /// \param b1  Field value in the first node [kG]
      /// \param z1  Second node z-coordinate [cm]
      /// \param b2  Field value in the first node [kG]
      /// \param z2  Third node z-coordinate [cm]
      FieldRegion_t MakeFieldRegion(const FieldValue_t& b0, const T& z0, const FieldValue_t& b1, const T& z1,
                                    const FieldValue_t& b2, const T& z2) const
      {
        return FieldRegion_t(fFieldType, b0, z0, b1, z1, b2, z2);
      }

     protected:
      /// \brief Default constructor
      FieldBase() = default;

      /// \brief  Copy constructor
      /// \tparam I  Underlying floating type of the source
      template<typename I>
      FieldBase(const FieldBase<I, EFieldMode::Intrpl>& other)
      {
        fvFieldSlices.clear();
        fvFieldSlices.reserve(other.fvFieldSlices.size());
        for (const auto& slice : other.fvFieldSlices) {
          fvFieldSlices.emplace_back(FieldSlice<T>(slice));
        }
      }

      /// \brief  Copy assignment operator
      FieldBase& operator=(const FieldBase& other) = default;

      SlicesContainer_t fvFieldSlices;          ///< Array of field slices
      EFieldType fFieldType{EFieldType::Null};  ///< Field type

     private:
      /// \brief Serialization function
      friend class boost::serialization::access;
      template<class Archive>
      void serialize(Archive& ar, const unsigned int)
      {
        ar& fvFieldSlices;
        ar& fFieldType;
      }
    };
  }  // namespace detail


  /// \class  Field
  /// \brief  Magnetic field manager class
  /// \tparam T  Underlying floating point type (float/double/fvec)
  template<typename T, EFieldMode FldMode>
  class alignas(VcMemAlign) Field : public detail::FieldBase<T, FldMode> {
    template<typename, EFieldMode>
    friend class Field;

   public:
    static constexpr EFieldMode FieldType = FldMode;

    using FieldRegion_t = FieldRegion<T, FldMode>;
    using Real_t        = T;

    /// \brief Default constructor
    Field() = default;

    /// \brief Copy constructor
    template<typename I>
    Field(const Field<I, FldMode>& other)
      : detail::FieldBase<T, FldMode>(other)
      , fPrimVertexField(FieldRegion<I, FldMode>(other.fPrimVertexField))
    {
    }

    /// \brief Destructor
    ~Field() = default;

    /// \brief Copy assignment operator
    Field& operator=(const Field& other) = default;

    /// \brief Gets field type
    EFieldType GetFieldType() const { return this->fFieldType; }

    /// \brief Gets field region near primary vertex
    const FieldRegion_t& GetPrimVertexField() const { return fPrimVertexField; }

    /// \brief Sets field type
    /// \param fieldType  Field type
    // TODO: SZh 21.08.2024:
    //   The concept of the field type depending on the z-regions should be developed. At the moment the full setup can
    //   have either EFieldType::Null (mCBM), or something else (CBM).
    void SetFieldType(EFieldType fieldType) { this->fFieldType = fieldType; }

    /// \brief Sets field region near primary vertex
    /// \param fld  Field region near vertex
    void SetPrimVertexField(const FieldRegion_t& primVertexField) { fPrimVertexField = primVertexField; }

    /// \brief String representation of the class
    /// \param indentLevel  Indent level of the string output
    /// \param verbose      Verbosity level
    std::string ToString(int indentLevel, int verbose) const;

   private:
    /// \brief Serialization function
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      ar& boost::serialization::base_object<detail::FieldBase<T, FldMode>>(*this);
      ar& fPrimVertexField;
    }

    FieldRegion_t fPrimVertexField;  ///< Field region near primary vertex (constant field is assumed)
  };


  /// \class FieldFactory
  /// \brief A helper class to instantiate a Field object
  class FieldFactory {
    /// \struct SliceRef
    /// \brief  A helper structure for field slices initialization
    struct SliceRef {
      double fHalfSizeX{defs::Undef<double>};  ///< Half-size of the slice in x-direction [cm]
      double fHalfSizeY{defs::Undef<double>};  ///< Half-size of the slice in y-direction [cm]
      double fRefZ{defs::Undef<double>};       ///< Reference z-position of the slice [cm]

      /// \brief Constructor
      /// \param halfX  Half-size of the slice in x-direction [cm]
      /// \param halfY  Half-size of the slice in y-direction [cm]
      /// \param refZ   Reference z-position of the slice [cm]
      SliceRef(double halfX, double halfY, double refZ) : fHalfSizeX(halfX), fHalfSizeY(halfY), fRefZ(refZ) {}

      /// \brief Comparision operator
      bool operator<(const SliceRef& r) const { return fRefZ < r.fRefZ; }
    };

   public:
    /// \brief Default constructor
    FieldFactory() = default;

    /// \brief Copy constructor
    FieldFactory(const FieldFactory&) = default;

    /// \brief Destructor
    ~FieldFactory() = default;

    /// \brief Copy assignment operator
    FieldFactory& operator=(const FieldFactory&) = default;

    /// \brief Adds a slice reference
    /// \param halfSizeX  Half-size of the slice in x-direction [cm]
    /// \param halfSizeY  Half-size of the slice in y-direction [cm]
    /// \param refZ       Reference z-position of the slice [cm]
    void AddSliceReference(double halfSizeX, double halfSizeY, double refZ);

    /// \brief Gets field type
    EFieldType GetFieldType() const { return fFieldType; }

    /// \brief  Create field manager
    /// \tparam T          Underlying floating-point data type
    /// \tparam OrigField  If orig field is used
    template<typename T, EFieldMode FldMode>
    Field<T, FldMode> MakeField() const;

    /// \brief Resets the instance
    void Reset() { *this = FieldFactory(); }

    /// \brief Sets magnetic field function
    /// \param fieldFn  Magnetic field function (KF-format)
    void SetFieldFunction(const FieldFn_t& fieldFn) { fFieldFn = fieldFn; }

    /// \brief Sets field type
    /// \param fieldType  Field type
    void SetFieldType(EFieldType fieldType) { fFieldType = fieldType; }

    /// \brief Sets target
    /// \param x  x-coordinate of the target position [cm]
    /// \param y  y-coordinate of the target position [cm]
    /// \param z  z-coordinate of the target position [cm]
    void SetTarget(double x, double y, double z) { fTarget = {x, y, z}; }

    /// \brief Sets a step for the primary vertex field region estimation
    /// \param step  A step between nodal points in z-axis direction [cm]
    void SetStep(double step = 2.5) { fTargetStep = step; }

   private:
    std::set<SliceRef> fSliceReferences;      ///< Set of slice references
    FieldFn_t fFieldFn{defs::ZeroFieldFn};    ///< Field function (x, y, z) [cm] -> (Bx, By, Bz) [kG]
    double fTargetStep{2.5};                  ///< Step between nodal points for the primary vertex field estimation
    EFieldType fFieldType{EFieldType::Null};  ///< Field type

    /// \brief Target position
    std::array<double, 3> fTarget{{defs::Undef<double>, defs::Undef<double>, defs::Undef<double>}};
  };

  // -------------------------------------------------------------------------------------------------------------------
  //
  template<typename T, EFieldMode FldMode>
  Field<T, FldMode> FieldFactory::MakeField() const
  {
    auto field = Field<T, FldMode>{};

    // Check initialization
    if (std::any_of(fTarget.begin(), fTarget.end(), [](double x) { return !utils::IsFinite(x); })) {
      throw std::logic_error("FieldFactory::MakeField: target is undefined");
    }
    if (!fSliceReferences.size()) {  // TODO: Remove requirement of slice references
      throw std::logic_error("FieldFactory::MakeField: no slice references were provided");
    }
    if (!fFieldFn) {
      throw std::logic_error("FieldFactory::CreateField: no field function is provided");
    }

    // Initialize the Field object
    field.SetFieldType(fFieldType);
    if constexpr (FldMode == EFieldMode::Orig) {
      field.SetFieldFunction(fFieldFn);
      field.SetPrimVertexField(field.MakeFieldRegion());
    }
    else if constexpr (FldMode == EFieldMode::Intrpl) {  // A version with the interpolated field
      for (const auto& sliceRef : fSliceReferences) {
        field.AddFieldSlice(FieldSlice<T>(fFieldFn, sliceRef.fHalfSizeX, sliceRef.fHalfSizeY, sliceRef.fRefZ));
      }
      {
        // PV field initialization
        constexpr size_t nNodes = 3;
        double z                = fTarget[2];
        std::array<FieldValue<T>, nNodes> fld;
        std::array<T, nNodes> pos;
        for (size_t iNode = 0; iNode < nNodes; ++iNode) {
          pos[iNode] = utils::simd::Cast<double, T>(z);
          fld[iNode] = GlobalField::GetFieldValue<double>(fFieldFn, fTarget[0], fTarget[1], z);
          z += fTargetStep;
        }
        field.SetPrimVertexField(field.MakeFieldRegion(fld[0], pos[0], fld[1], pos[1], fld[2], pos[2]));
      }
    }
    return field;
  }
}  // namespace cbm::algo::kf
