/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   KfFieldRegion.h
/// \brief  Magnetic flux density interpolation along the track vs. z-coordinate (header)
/// \author Sergei Zharko <s.zharko@gsi.de>
/// \since  19.07.2024
///

#pragma once

#include "KfDefs.h"
#include "KfFieldValue.h"
#include "KfMath.h"
#include "KfUtils.h"

#include <boost/serialization/access.hpp>
#include <boost/serialization/array.hpp>
#include <boost/serialization/base_object.hpp>

#include <array>
#include <string>
#include <type_traits>

namespace cbm::algo::kf
{

  /// \class GlobalField
  /// \brief Handler for the global magnetic field related functions
  class GlobalField {
   public:
    /// \brief Global variable to store the fielf funciton (x,y,z)->(Bx,By,Bz)
    static FieldFn_t gOriginalField;

    /// \brief Undefined magnetic field function
    static std::tuple<double, double, double> UndefField(double, double, double);

    /// \brief Sets global field function
    /// \param fn  Field function of the FairRoot format
    static void SetFieldFunction(const FieldFn_t& fn) { gOriginalField = fn; }

    /// \brief Creates a field value object in a spactial point of a generic floating-point type
    /// \tparam T       Floating point data-type of the involved variables
    /// \param fieldFn  Field function (x,y,z)->(Bx,By,Bz)
    /// \param x        The point x-coordinate [cm]
    /// \param y        The point y-coordinate [cm]
    /// \param z        The point z-coordinate [cm]
    /// \return  FieldValue<T> object
    template<typename T>
    [[gnu::always_inline]] static FieldValue<T> GetFieldValue(const FieldFn_t& fieldFn, const T& x, const T& y,
                                                              const T& z);
  };

  namespace detail
  {
    /// \class  FieldRegionBase
    /// \brief  Base class of the FieldRegion class (primary template)
    /// \tparam T          Underlying floating point type (float/double/fvec)
    /// \tparam OrigField  When true, the original CBM field function is used. When false, the polynomial
    ///                    interpolation is used instead.
    template<typename T, EFieldMode FldMode>
    class alignas(VcMemAlign) FieldRegionBase {
    };


    /// \class  FieldRegionBase<T, EFieldMode::Orig>
    /// \brief  Data members and specific methods of the FieldRegion with the original magnetic field
    /// \tparam T  Underlying floating point type (float/double/fvec)
    template<typename T>
    class alignas(VcMemAlign) FieldRegionBase<T, EFieldMode::Orig> {
      template<typename, EFieldMode>
      friend class FieldRegionBase;

     public:
      using FieldValue_t = FieldValue<T>;

      /// \brief Default constructor
      FieldRegionBase() = default;

      /// \brief Constructor
      FieldRegionBase(const FieldFn_t& fieldFn) : fFieldFn(fieldFn) {}

      /// \brief  Copy constructor
      /// \tparam I  Underlying floating point type of the source
      template<typename I>
      FieldRegionBase(const FieldRegionBase<I, EFieldMode::Orig>& other) : fFieldFn(other.fFieldFn)
      {
      }

      /// \brief Destructor
      ~FieldRegionBase() = default;

      /// \brief Copy assignment operator
      FieldRegionBase& operator=(const FieldRegionBase& other) = default;

     protected:
      FieldFn_t fFieldFn{defs::ZeroFieldFn};  ///< Field function: (x,y,z) [cm] -> (Bx,By,Bz) [kG]

     private:
      /// \brief Serialization function
      friend class boost::serialization::access;
      template<class Archive>
      void serialize(Archive&, const unsigned int)
      {
      }
    };


    /// \class  FieldRegionBase<T, Intrpl>
    /// \brief  Data members and specific mehtods of the FieldRegion class with the interpolated magnetic field
    /// \tparam T  Underlying floating  point type (float/double/fvec)
    template<typename T>
    class alignas(VcMemAlign) FieldRegionBase<T, EFieldMode::Intrpl> {
      template<typename, EFieldMode>
      friend class FieldRegionBase;

      using CoeffArray_t = std::array<std::array<T, 3>, 3>;  ///< Approximation coefficients array

     public:
      using FieldValue_t = FieldValue<T>;

      /// \brief Default constructor
      FieldRegionBase() = default;

      /// \brief Constructor from parameters
      /// \param b0  Field value in the first node [kG]
      /// \param z0  First node z-coordinate [cm]
      /// \param b1  Field value in the first node [kG]
      /// \param z1  Second node z-coordinate [cm]
      /// \param b2  Field value in the first node [kG]
      /// \param z2  Third node z-coordinate [cm]
      FieldRegionBase(const FieldValue_t& b0, const T& z0, const FieldValue_t& b1, const T& z1, const FieldValue_t& b2,
                      const T& z2)
      {
        this->Set(b0, z0, b1, z1, b2, z2);
      }

      /// \brief Copy constructor
      template<typename I>
      FieldRegionBase(const FieldRegionBase<I, EFieldMode::Intrpl>& other)
        : fZfirst(utils::simd::Cast<I, T>(other.fZfirst))
      {
        for (size_t iDim = 0; iDim < fCoeff.size(); ++iDim) {
          for (size_t iPow = 0; iPow < fCoeff[iDim].size(); ++iPow) {
            fCoeff[iDim][iPow] = utils::simd::Cast<I, T>(other.fCoeff[iDim][iPow]);
          }
        }
      }

      /// \brief Destructor
      ~FieldRegionBase() = default;

      /// \brief Copy assignment operator
      FieldRegionBase& operator=(const FieldRegionBase& other) = default;

      /// \brief Gets the first z-coordinate of the interpolation
      const T& GetZfirst() const { return fZfirst; }

      /// \brief Interpolates the field by three nodal points with the second order polynomial
      /// \note  The interpolation is carried out vs. z-coordinates of the space points of the field values
      /// \param b0  Field value at z0 [kG]
      /// \param z0  First nodal point in z-axis direction [cm]
      /// \param b1  Field value at z1 [kG]
      /// \param z1  Second nodal point in z-axis direction [cm]
      /// \param b2  Field value at z2 [kG]
      /// \param z2  Third nodal point in z-axis direction [cm]
      void Set(const FieldValue<T>& b0, const T& z0, const FieldValue<T>& b1, const T& z1, const FieldValue<T>& b2,
               const T& z2);

      /// \brief Shifts the coefficients to another first z-coordinate
      /// \param z  new z-coordinate [cm]
      void Shift(const T& z);

     protected:
      CoeffArray_t fCoeff{{T(0.)}};  ///< Approximation coefficients
      T fZfirst{Literal_t<T>()};     ///< Reference z-coordinate of the field region

     private:
      /// Serialization function
      friend class boost::serialization::access;
      template<class Archive>
      void serialize(Archive& ar, const unsigned int)
      {
        ar& fCoeff;
        ar& fZfirst;
      }
    };
  }  // namespace detail


  /// \class  FieldRegion
  /// \brief  Magnetic field region, corresponding to a hit triplet
  /// \tparam T        Underlying data-type
  /// \tparam FldMode  Field type
  template<typename T, EFieldMode FldMode>
  class alignas(VcMemAlign) FieldRegion : public detail::FieldRegionBase<T, FldMode> {
    template<typename, EFieldMode>
    friend class FieldRegion;

   public:
    static constexpr EFieldMode FieldType = FldMode;

    using FieldValue_t = FieldValue<T>;
    using Real_t       = T;

    /// \brief Default constructor
    FieldRegion() = default;

    /// \brief  Copy constructor
    /// \tparam I  Underlying floating point type of the source
    template<typename I>
    FieldRegion(const FieldRegion<I, FldMode>& other) : detail::FieldRegionBase<T, FldMode>(other)
    {
    }

    /// \brief  Constructor from the arguments
    /// \param  fieldType  Type of the field
    /// \param  args       Arguments, which have to be passed to the base class constructors
    template<typename... Args>
    FieldRegion(EFieldType fieldType, Args... args)
      : detail::FieldRegionBase<T, FldMode>(args...)
      , fFieldType(fieldType)
    {
    }

    /// \brief Destructor
    ~FieldRegion() = default;

    /// \brief Copy assignment operator
    FieldRegion& operator=(const FieldRegion& other) = default;

    /// \brief Gets the field value in a spatial point on the track
    /// \param x  x-coordinate of the point [cm]
    /// \param y  y-coordinate of the point [cm]
    /// \param z  z-coordinate of the point [cm]
    /// \note  The x and y coordinates are ignored, if the interpolated field is used
    FieldValue_t Get(const T& x, const T& y, const T& z) const;

    /// \brief Gets the field type
    EFieldType GetFieldType() const { return fFieldType; }

    /// \brief String representation of the class content
    /// \param intdentLevel  Indent level
    /// \param verbose       Verbosity level
    std::string ToString(int indentLevel = 0, int verbose = 1) const;

   private:
    /// \brief Serialization method
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      using boost::serialization::base_object;
      ar& base_object<detail::FieldRegionBase<T, FldMode>>(*this);
      ar& fFieldType;
    }

    EFieldType fFieldType{EFieldType::Null};  ///< Field type in a given region
  };


  // -------------------------------------------------------------------------------------------------------------------
  //
  template<typename T>
  inline FieldValue<T> GlobalField::GetFieldValue(const FieldFn_t& fieldFn, const T& x, const T& y, const T& z)
  {
    using utils::simd::Cast;
    FieldValue<T> B;
    for (size_t i = 0; i < utils::simd::Size<T>(); ++i) {
      auto [bx, by, bz] = fieldFn(Cast<T, double>(x, i), Cast<T, double>(y, i), Cast<T, double>(z, i));
      B.SetSimdEntry(bx, by, bz, i);
    }
    return B;
  }

}  // namespace cbm::algo::kf
