/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfSetup.h
/// @brief  Setup representation for the Kalman-filter framework (header)
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once  // include this header only once per compilation unit

#include "KfDefs.h"
#include "KfField.h"
#include "KfMaterialMap.h"
#include "KfTarget.h"
#include "KfVector.h"

#include <boost/serialization/access.hpp>
#include <boost/serialization/split_free.hpp>

#include <fstream>
#include <optional>
#include <sstream>
#include <string>
#include <vector>

namespace cbm::algo::kf
{
  /// \class  Setup
  /// \brief  KF-framework representation of the detector setup
  /// \tparam T  Underlying floating-point data type
  template<typename T>
  class alignas(VcMemAlign) Setup {
    friend class boost::serialization::access;  // Boost serializer methods
    template<typename>
    friend class Setup;


   public:
    /// \brief Default constructor
    Setup() = default;

    /// \brief Destructor
    ~Setup() = default;

    /// \brief  Copy constructor
    /// \tparam I  Underlying floating-point data type of the source
    template<typename I>
    Setup(const Setup<I>& other)
      : fvMaterialLayers(other.fvMaterialLayers)
      , fIntrplField(other.fIntrplField)
      , foOrigField(other.foOrigField.has_value() ? std::make_optional(*other.foOrigField) : std::nullopt)
      , fTarget(other.fTarget)
    {
    }

    /// \brief Move constructor
    //Setup(Setup&&) noexcept;

    /// \brief  Copy assignment operator
    Setup& operator=(const Setup& other) = default;

    /// \brief Move assignment operator
    //Setup& operator=(Setup&&) noexcept;

    /// \brief Sets material layer
    /// \param map     Material map
    void AddMaterial(const MaterialMap& map) { fvMaterialLayers.emplace_back(map); }

    /// \brief Makes an instance of the field depending on the template parameter
    /// \tparam FldMode           Field mode of the returned instance
    /// \throw  std::logic_error  If the particular field member is undefined (nullopt)
    template<EFieldMode FldMode>
    const Field<T, FldMode>& GetField() const;

    /// \brief Gets material layer
    /// \param iLayer  Index of layer
    const MaterialMap& GetMaterial(int iLayer) const { return fvMaterialLayers[iLayer]; }

    /// \brief Gets number of defined material layers
    int GetNofMaterialLayers() const { return static_cast<int>(fvMaterialLayers.size()); }

    /// \brief Gets target
    const Target<T>& GetTarget() const { return fTarget; }

    /// \brief String representation of the class contents
    /// \param verbosity    A verbose level for output
    /// \param indentLevel  Indent level of the string output
    std::string ToString(int verbosity = 1, int indentLevel = 0) const;

    /// \brief Sets field
    template<EFieldMode FldMode>
    void SetField(const Field<double, FldMode>& field);

    /// \brief Sets target (moves from the source)
    /// \param target  Target instance
    void SetTarget(const Target<double>& target) { fTarget = target; }

   private:
    /// \brief BOOST serialization load method
    template<class Archive>
    void load(Archive& ar, const unsigned int /*version*/)
    {
      ar >> fvMaterialLayers;
      ar >> fvMaterialLayerIds;
      ar >> fTarget;
      ar >> fIntrplField;
      foOrigField = std::nullopt;

      // NOTE: The original field cannot be serialized, because it contains a pointer to the external magnetic field
      //       function.
    }

    template<class Archive>
    void save(Archive& ar, const unsigned int /*version*/) const
    {
      ar << fvMaterialLayers;
      ar << fvMaterialLayerIds;
      ar << fTarget;
      ar << fIntrplField;
    }

    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
    {
      boost::serialization::split_member(ar, *this, version);
    }

    std::vector<MaterialMap> fvMaterialLayers = {};  ///< Container of the material maps
    std::vector<int> fvMaterialLayerIds       = {};  ///< Map of the layer ids: original -> actual
    Field<T, EFieldMode::Intrpl> fIntrplField;       ///< Interpolated field (NOTE: maybe make optional)
    std::optional<Field<T, EFieldMode::Orig>> foOrigField{std::nullopt};  ///< Optional original field
    Target<T> fTarget;                                                    ///< Target layer
  };

  // -------------------------------------------------------------------------------------------------------------------
  //
  template<typename T>
  template<EFieldMode FldMode>
  const Field<T, FldMode>& Setup<T>::GetField() const
  {
    if constexpr (FldMode == EFieldMode::Orig) {
      if (foOrigField) {
        return *foOrigField;
      }
      else {
        throw std::logic_error("kf::Setup::MakeField(): attempt to create a Field instance in the original field mode "
                               "in the unsuitable scenario (for example, in the online reconstruction code). Please, "
                               "use the interpolated magnetic field calling Setup::MakeField<T, EFieldMode::Intrpl>() "
                               "instead.");
      }
    }
    else if constexpr (FldMode == EFieldMode::Intrpl) {
      return fIntrplField;
    }
  }

  // -------------------------------------------------------------------------------------------------------------------
  //
  template<typename T>
  template<EFieldMode FldMode>
  void Setup<T>::SetField(const Field<double, FldMode>& field)
  {
    if constexpr (FldMode == EFieldMode::Intrpl) {
      fIntrplField = field;
    }
    else if constexpr (FldMode == EFieldMode::Orig) {
      foOrigField = std::make_optional(field);
    }
  }
}  // namespace cbm::algo::kf
