/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfSetupInitializer.h
/// @brief  A base KF-Setup initialization class (header)
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once  // include this header only once per compilation unit

#include "KfSetup.h"

#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>

#include <fstream>
#include <optional>
#include <sstream>
#include <string>
#include <vector>

namespace cbm::algo::kf
{
  /// \class  SetupInitializer
  /// \brief  Creates a valid initialized Setup instance
  class SetupInitializer {
   public:
    /// \brief Default constructor
    SetupInitializer() = default;

    /// \brief Copy constructor
    SetupInitializer(const SetupInitializer&) = delete;

    /// \brief Move constructor
    SetupInitializer(SetupInitializer&&) = delete;

    /// \brief Destructor
    virtual ~SetupInitializer() = default;

    /// \brief Copy assignment operator
    SetupInitializer& operator=(const SetupInitializer&) = delete;

    /// \brief Move assignment operator
    SetupInitializer& operator=(SetupInitializer&&) = delete;

    /// \brief Adds magnetic field slice reference
    /// \param halfSizeX  Half-size of the slice in x-direction [cm]
    /// \param halfSizeY  Half-size of the slice in y-direction [cm]
    /// \param refZ       Reference z-position of the slice [cm]
    void AddFieldSliceReference(double halfSizeX, double halfSizeY, double refZ)
    {
      fFieldFactory.AddSliceReference(halfSizeX, halfSizeY, refZ);
    }

    /// \brief Adds material layer
    /// \param material  Material layer
    void AddMaterial(const MaterialMap& material);

    /// \brief  Virtual function, which initializes a KF-setup
    /// \return true  Initialization was successful
    /// \return false Initialization failed
    ///
    /// This function can be re-implemented for a particular use-case...
    virtual bool InitProperties() { return false; }

    /// \brief Creates a setup instance
    /// \param bProvideOrigField  Flag: true - defines optional variable of the original field
    template<typename T>
    Setup<T> MakeSetup(bool bProvideOrigField = true) const;

    /// \brief Resets the instance
    void Reset();

    /// \brief Sets magnetic field function
    /// \param fieldFn    Magnetic field function
    /// \param fieldType  Magnetic field type
    void SetFieldFunction(const FieldFn_t& fieldFn, EFieldType fieldType)
    {
      fFieldFactory.SetFieldFunction(fieldFn);
      fFieldFactory.SetFieldType(fieldType);
    }

    /// \brief Sets target initialization properties
    /// \param x               Target x-coordinate [cm]
    /// \param y               Target y-coordinate [cm]
    /// \param z               Target z-coordinate [cm]
    /// \param fieldInitStep   Step between the nodal points in the target field region initialization [cm]
    /// \param material        Target material layer
    void SetTargetProperties(double x, double y, double z, double fieldInitStep, const MaterialMap& material);

    /// \brief  Stores a serialized setup to a file
    /// \param  setup     A reference to the Setup (NOTE: a double-Setup must be passed explicitly)
    /// \param  fileName  Output file name
    ///
    /// Only a Setup with T = double can be serialized, so if a particular setup instance has another floating-point
    /// type, it will be converted to the double-version. In the serialization only the interpolated magnetic field
    /// variant is stored, the original field version will be set to nullopt during the loading from the file.
    static void Store(const Setup<double>& setup, const std::string& fileName);

    /// \brief  Loads a serialized setup from a file
    /// \tparam T         Underlying floating-point type for the output setup object
    /// \param  fileName  Input file name
    /// \throw  std::runtime_error  If the fileName cannot be opened
    template<typename T>
    static Setup<T> Load(const std::string& fileName);

   private:
    std::set<MaterialMap> fMaterialLayers{};  ///< Set of material maps
    FieldFactory fFieldFactory;               ///< Instance of field factory
    Target<double> fTarget;                   ///< Field to keep target properties
    bool fbPropertiesInitialized{false};      ///< Flag: if all the necessary fields are initialized
  };


  // ********************************
  // ** Template method definition **
  // ********************************

  // ---------------------------------------------------------------------------------------------------------------------
  //
  template<typename T>
  Setup<T> SetupInitializer::MakeSetup(bool bProvideOrigField) const
  {
    Setup<T> setup;
    // Target initialization
    setup.SetTarget(fTarget);

    // Field initialization
    setup.SetField(fFieldFactory.MakeField<double, EFieldMode::Intrpl>());
    if (bProvideOrigField) {
      setup.SetField(fFieldFactory.MakeField<double, EFieldMode::Orig>());
    }

    // Material layers initialization
    for (const auto& material : fMaterialLayers) {
      setup.AddMaterial(material);
    }

    return setup;
  }

  // -------------------------------------------------------------------------------------------------------------------
  //
  template<typename T>
  Setup<T> SetupInitializer::Load(const std::string& fileName)
  {
    Setup<double> setup;
    std::ifstream ifs(fileName, std::ios::binary);
    if (!ifs) {
      std::stringstream msg;
      msg << "kf::SetupInitializer::Load: intput setup file \"" << fileName << "\" was not found";
      throw std::runtime_error(msg.str());
    }
    try {
      boost::archive::binary_iarchive ia(ifs);
      ia >> setup;
    }
    catch (const std::exception& err) {
      std::stringstream msg;
      msg << "kf::SetupInitializer::Load: input setup file \"" << fileName
          << "\" has inconsistent format or was "
             "corrupted. The exception message: "
          << err.what();
      throw std::runtime_error(msg.str());
    }
    return Setup<T>(setup);
  }
}  // namespace cbm::algo::kf
